@@ -4,19 +4,12 @@ use strict;
 package Devel::hdb::DB;
 
 use Scalar::Util;
+use IO::File;
 
 use Devel::hdb::DB::Eval;
 
 package DB;
 
-# NOTE: Look into trapping $SIG{__DIE__} se we can report
-# untrapped exceptions back to the debugger.
-# inside the handler, note the value for $^S:
-# undef - died while parsing something
-# 1 - died while executing an eval
-# 0 - Died not inside an eval
-# We could re-throw the die if $^S is 1
-
 use vars qw( %dbline @dbline );
 
 our($stack_depth,
@@ -24,6 +17,7 @@ our($stack_depth,
     $signal,
     $trace,
     $debugger_disabled,
+    $no_stopping,
     $step_over_depth,
     $dbobj,
     $ready,
@@ -36,6 +30,8 @@ our($stack_depth,
     $eval_string,
     @AUTOLOAD_names,
     $sub,
+    $uncaught_exception,
+    $input_trace,
 );
 
 BEGIN {
@@ -43,6 +39,7 @@ BEGIN {
     $single         = 0;
     $trace          = 0;
     $debugger_disabled = 0;
+    $no_stopping    = 0;
     $step_over_depth = undef;
     $dbobj          = undef;
     $ready          = 0;
@@ -78,8 +75,73 @@ sub restore {
     ( $@, $!, $^E, $,, $/, $\, $^W ) = @saved;
 }
 
+sub _line_offset_for_sub {
+    my($line, $subroutine) = @_;
+    no warnings 'uninitialized';
+    if ($DB::sub{$subroutine} =~ m/(\d+)\-\d+$/) {
+        return $line - $1;
+    } else {
+        return undef;
+    }
+}
+
+sub _trace_report_line {
+    my($package, $filename, $line, $subroutine) = @_;
+
+    my $location;
+    if (my $offset = _line_offset_for_sub($line, $subroutine)) {
+        $location = "${subroutine}+${offset}";
+    } else {
+        $location = "${filename}:${line}";
+    }
+
+    return join("\t", $location, $package, $filename, $line, $subroutine);
+}
+
+sub input_trace_file {
+    my($class, $file, $cb) = @_;
+    my $fh = IO::File->new($file, 'r');
+    unless ($fh) {
+        warn "Can't open trace file $file for reading: $!";
+    }
+    $input_trace = sub {
+        my($package, $filename, $line, $subroutine) = @_;
+
+        my @line = split("\t", $fh->getline);
+        my($offset, $expected_sub, $expected_offset, $should_stop);
+
+        if (($expected_sub, $expected_offset) = $line[0] =~ m/(.*?)\+(\d+)/) {
+            $offset = _line_offset_for_sub($line, $subroutine);
+            $should_stop = ($expected_sub ne $subroutine or $expected_offset != $offset);
+        } else {
+            my($file, $fileline) = $line[0] =~ m/(.*)?\:(\d+)/;
+            $should_stop = ($file ne $filename or $fileline != $line);
+        }
+        if ($should_stop) {
+            my $diff_data = {
+                'package'   => $package,
+                filename    => $filename,
+                line        => $line,
+                subroutine  => $subroutine,
+                sub_offset  => _line_offset_for_sub($line, $subroutine),
+            };
+            @$diff_data{'expected_package', 'expected_filename', 'expected_line',
+                        'expected_subroutine','expected_sub_offset'}
+                = (@line[1, 2, 3, 4], $expected_offset);
+            $cb->($diff_data);
+            undef $input_trace; # It's likely _every_ line will now be different
+        }
+        return $should_stop;
+    };
+    $trace = 1;
+}
+
 sub is_breakpoint {
-    my($package, $filename, $line) = @_;
+    my($package, $filename, $line, $subroutine) = @_;
+
+    if ($input_trace && $input_trace->($package, $filename, $line, $subroutine)) {
+        return 1;
+    }
 
     if ($single and defined($step_over_depth) and $step_over_depth < $stack_depth) {
         # This is from a step-over
@@ -138,6 +200,43 @@ BEGIN {
     };
 };
 
+# NOTE: Look into trapping $SIG{__DIE__} se we can report
+# untrapped exceptions back to the debugger.
+# inside the handler, note the value for $^S:
+# undef - died while parsing something
+# 1 - died while executing an eval
+# 0 - Died not inside an eval
+# We could re-throw the die if $^S is 1
+$SIG{__DIE__} = sub {
+    if (defined($^S) && $^S == 0) {
+        my $exception = $_[0];
+        # It's interesting to note that if we pass an arg to caller() to
+        # find out the offending subroutine name, then the line reported
+        # changes.  Instead of reporting the line the exception occured
+        # (which it correctly does with no args), it returns the line which
+        # called the function which threw the exception.
+        # We'll work around it by calling it twice
+        my($package, $filename, undef, $subname) = caller(1);
+        my(undef, undef, $line, undef) = caller(0);
+        $subname = 'MAIN' unless defined($subname);
+        $uncaught_exception = {
+            'package'   => $package,
+            line        => $line,
+            filename    => $filename,
+            exception   => $exception,
+            subroutine  => $subname,
+        };
+        # After we fall off the end, the interpreter will try and exit,
+        # triggering the END block that calls DB::fake::at_exit()
+    }
+};
+
+
+sub disable_stopping {
+    my $class = shift;
+    $no_stopping = shift;
+}
+
 sub DB {
     return if (!$ready or $debugger_disabled);
 
@@ -158,7 +257,16 @@ sub DB {
         &eval;
     }
 
-    if (! is_breakpoint($package, $filename, $line)) {
+    my(undef, undef, undef, $subroutine) = caller(1);
+    $subroutine ||= 'MAIN';
+    if ($trace && ! $input_trace) {
+        my $fh = (ref($trace) and $trace->can('print')) ? $trace : \*STDERR;
+        $fh->print( _trace_report_line($package, $filename, $line, $subroutine), "\n");
+    }
+
+    return if $no_stopping;
+
+    if (! is_breakpoint($package, $filename, $line, $subroutine)) {
         return;
     }
     $step_over_depth = undef;
@@ -367,13 +475,6 @@ sub long_call {
     return $DB::long_call;
 }
 
-# FIXME: I think the keys for %DB::sub is fully qualified
-# sub names, like Package::Subpkg::subname
-# values are "filename:startline-endline"
-sub subroutines {
-
-}
-
 sub user_requested_exit {
     $user_requested_exit = 1;
 }
@@ -399,7 +500,8 @@ sub disable_debugger {
 
 
 END {
-    print "Debugged program pid $$ terminated with exit code $?\n";
+    $trace = 0;
+
     return if $debugger_disabled;
 
     $single=0;
@@ -409,7 +511,7 @@ END {
     if ($user_requested_exit) {
         Devel::hdb::App->get()->notify_program_exit();
     } else {
-        Devel::hdb::App->get()->notify_program_terminated($?);
+        Devel::hdb::App->get()->notify_program_terminated($?, $uncaught_exception);
         # These two will trigger DB::DB and the event loop
         $in_debugger = 0;
         $single=1;
